#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# _____________________________________________________________________
#
# Imports 
#   * imported by import order.
# 
from dev0s.classes import utils
from dev0s.classes.defaults.exceptions import Exceptions as exceptions
from dev0s.classes.defaults.color import color, symbol
from dev0s.classes import console
from dev0s.classes.system.env import env
from dev0s.classes.defaults import files
from dev0s.classes.response import response
from dev0s.classes.defaults.defaults import defaults
from dev0s.classes.defaults import objects
from dev0s.classes import database
from dev0s.classes import cli
from dev0s.classes.requests import requests
from dev0s.classes import code
from dev0s.classes.network import network
from dev0s.classes.system import System as system
from dev0s.classes.encryption import Encryption as encryption

# _____________________________________________________________________
#
# Universal:
#   * universal package imports (do not edit).
#

# source, base & version.
source = files.Directory(files.gfp.base(__file__))
base = files.Directory(source.fp.base())
version = code.Version(files.Files.load(source.join(".version")))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.defaults.exceptions import Exceptions
from dev0s.classes.defaults.color import color, symbol
from dev0s.classes.defaults.files import *
from dev0s.classes.response import response as _response_

# the env object class.
class Env(object):
	def __init__(self):
		
		# docs.
		DOCS = {
			"module":"dev0s.system.env", 
			"initialized":True,
			"description":[], }

		#
	
	# fill with variables.
	def fill(self, string):
		c = 0
		for i in range(string.count("$")):
			c += 1
			if string[0] == "$" and c == 1:
				variable_id, _ = String(string.split("$")[1]).before_after_first_occurence(slicer=["/","+",">","<","?",";",":","|","\\","]","[","{","}","=","+","(",")","!","@","#","$","%","^","&","*","±","~","`","'",'"'," "], include_after=True)
			else:
				variable_id, _ = String(string.split("$")[c]).before_after_first_occurence(slicer=["/","+",">","<","?",";",":","|","\\","]","[","{","}","=","+","(",")","!","@","#","$","%","^","&","*","±","~","`","'",'"'," "], include_after=True)
			for i in ["/","+",">","<","?",";",":","|","\\","]","[","{","}","=","+","(",")","!","@","#","$","%","^","&","*","±","~","`","'",'"'," "]:
				variable_id = variable_id.replace(i, "")
			variable = os.environ.get(variable_id)
			if variable == None: raise ValueError(f"Required environment variable {variable_id} is undefined.")
			string = string.replace(f"${variable_id}", variable)
		string = string.replace("~",HOME)
		return string
	
	# import env.
	def import_(self, env=None):
		if env == None: return _response_.error("Define parameter: env.")
		dictionary = {}
		if isinstance(env, (str, String, FilePath)):
			env = str(env)
			if ".json" in env or env == "json":
				try:
					dictionary = Files.load(env, format="json")
				except FileNotFoundError:
					dictionary = {}
			elif ".sh" in env or env == "bash":
				return _response_.error("Coming soon")
		elif isinstance(env, (dict, Dictionary)):
			dictionary = env
		else: return _response_.error(f"Invalid usage, env requires to be a [str, String, FilePath, dict, Dictionary], not [{env.__class__.__name__}].")
		for key,value in dictionary.items():
			os.environ[str(key)] = str(value)
		return _response_.success(f"Successfully imported {len(dictionary)} env variables.")
		#
	
	# export env.
	def export(self, 
		# the environment to export (dict).
		env=None, 
		# the export path (str) or paths (list).
		# the paths must have .json / .sh extension or be named 'json' / 'bash' when parameter [format] is undefined.
		export=None,
		# the export format (str) (leave None to be detected by parameter [export]).
		format=None,
	):
		if env == None: return _response_.error("Define parameter: env.")
		dictionary = {}
		if isinstance(env, (dict, Dictionary)):
			dictionary = env
		else: return _response_.error(f"Invalid usage, env requires to be a [str, String, FilePath, dict, Dictionary], not [{env.__class__.__name__}].")
		for key,value in dictionary.items():
			os.environ[str(key)] = str(value)
		if export != None:
			if isinstance(export, (str,String,FilePath)):
				exports = [str(export)]
			elif not isinstance(export, (list, Array)):
				raise Exceptions.InvalidUsage(f"<dev0s.system.env>: Parameter export requires to be a [str, list, FilePath], not [{export.__class__.__name__}].")
			else:
				exports = export
			c = 0
			for export in exports:
				if isinstance(format, (list, Array)):
					try:
						l_format = format[c]
					except:
						raise Exceptions.InvalidUsage(f"<dev0s.system.env>: Parameter format [list] does not contain index [{c}] for export [{export}].")
				else:
					l_format = format
				if (l_format == "json") or ".json" in export or gfp.name(export) == "json":
					format = "json"
					try:
						exported = Files.load(export, format="json")
					except FileNotFoundError:
						exported = {}
				elif (l_format == "bash") or ".sh" in export or gfp.name(export) == "bash":
					format = "bash"
					try:
						exported = Files.load(export, format="str")
					except FileNotFoundError:
						exported = ""
					while True:
						if "\n\n" in exported: exported = exported.replace("\n\n","\n")
						elif len(exported) > 0 and String(exported).last("\n") == "\n": exported = str(String(exported).remove_last("\n"))
						else: break
				else:
					raise Exceptions.InvalidUsage(f"Export file [{export}] must contain an .json / .sh extension or must be named 'bash' / 'json'.")
				for key,value in dictionary.items():
					if format == "json":
						exported[str(key)] = value
					elif format == "bash":
						if f'export {key.upper()}="' in exported:
							l = ""
							for line in exported.split("\n"):
								if f'export {key.upper()}="' in line:
									l += f'export {key.upper()}="{value}"\n'
								else:
									l += line+"\n"
							exported = l
						else:
							exported += '\n'+f'export {key.upper()}="{value}"\n'
				if format == "json":
					Files.save(export, exported, format="json")
				elif format == "bash":
					while True:
						if "\n\n" in exported: exported = exported.replace("\n\n","\n")
						elif len(exported) > 0 and String(exported).last("\n") == "\n": exported = str(String(exported).remove_last("\n"))
						else: break
					Files.save(export, exported, format="str")
				c += 1
		return _response_.success(f"Successfully exported {len(dictionary)} env variables.")
		#
	
	# get.
	def get(self, id, default=None, format="str"):
		if format in [str, "str", "string"]:
			return self.get_string(id, default=default)
		elif format in [bool, "bool", "boolean"]:
			return self.get_boolean(id, default=default)
		elif format in [int, float, "int", "float", "integer", "double"]:
			return self.get_integer(id, default=default)
		elif list in [list, "list", "array"]:
			return self.get_array(id, default=default)
		elif dict in [dict, "dict", "dictionary", "json"]:
			return self.get_dictionary(id, default=default)
		else: raise ValueError(f"Unknown format: {format}.")
	def get_string(self, id, default=None):
		e = os.environ.get(id)
		if e == None: 
			return default
		else:
			return e
	def get_boolean(self, id, default=None):
		e = os.environ.get(id)
		if e == None: 
			return default
		elif e in [True, "True", "true", "TRUE"]:
			return True
		else: 
			return False
	def get_integer(self, id, default=None):
		e = os.environ.get(id)
		try:
			if "." in str(e):
				return float(e)
			else:
				return int(e)
		except:
			return default
	def get_array(self, id, default=None):
		e = os.environ.get(id)
		if e == None: 
			return default
		else: 
			a = []
			for i in e.split(","):
				a.append(self.__remove_whitespace__(i))
			return a
	def get_tuple(self, id, default=None):
		e = os.environ.get(id)
		if e == None: 
			return default
		else: 
			a = []
			for i in e.split(","):
				a.append(self.__remove_whitespace__(i))
			return tuple(a)
	def get_dictionary(self, id, default=None):
		e = os.environ.get(id)
		if e == None: 
			return default
		else: 
			a = []
			for i in e.split(","):
				a.append(self.__remove_whitespace__(i))
			d = {}
			for i in a:
				try:
					key,value = i.split(":")
					d[self.__remove_whitespace__(key)] = self.__remove_whitespace__(value)
				except: a=1
			return tuple(a)
	
	# set.
	def set(self, id, value, format="unknown"):
		if format == "unknown": format = Formats.get(value, serialize=True)
		if format in [str, "str", "string"]:
			self.set_string(id, value)
		elif format in [bool, "bool", "boolean"]:
			self.set_boolean(id, value)
		elif format in [int, float, "int", "float", "integer", "double"]:
			self.set_integer(id, value)
		elif list in [list, "list", "array"]:
			self.set_array(id, value)
		elif dict in [dict, "dict", "dictionary", "json"]:
			self.set_dictionary(id, value)
		else: raise ValueError(f"Unknown format: {format}.")
	def set_string(self, id, value):
		os.environ[id] = value
	def set_boolean(self, id, value):
		if e in [True, "True", "true", "TRUE"]:
			os.environ[id] = "true"
		else: 
			os.environ[id] = "false"
	def set_integer(self, id, value):
		os.environ[id] = str(value)
	def set_array(self, id, value):
		os.environ[id] = Array(value).string(joiner=",")
	def set_tuple(self, id, value):
		os.environ[id] = Array(value).string(joiner=",")
	def set_dictionary(self, id, value, subkey=""):
		for key, value in value.items():
			if subkey == "": lkey = key
			else: lkey = f"{subkey}_{key}"
			if isinstance(value, dict):
				self.set_dictionary(key, value, subkey=lkey)
			else: 
				self.set(lkey, value, format="unknown")

	# support item assignment.
	def __getitem__(self, key):
		value = self.get(id)
		if value in ["None", "none"]: value = None
		elif value in ["True", "true", "TRUE"]: value = True
		elif value in ["False", "false", "FALSE"]: value = False
		else:
			if "." in str(key):
				try: value = float(value)
				except: a=1
			else:
				try: value = int(value)
				except: a=1
		return value
	def __setitem__(self, key, value):
		self.set(key, str(value), format="str")
	def __delitem__(self, key):
		self.set(key, "")

	# system functions.
	def __remove_whitespace__(self, string):
		if isinstance(string, str):
			for attempts in range(101):
				if string[0] == " ":
					string = string[1:]
				elif string[len(string)-1] == " ":
					string = string[:-1]
				else: break
		return string

	#
	
# initialized classes.
env = Env()




# imports
from dev0s.classes.defaults.exceptions import Exceptions
from dev0s.classes.defaults.color import color
from dev0s.classes.defaults.files import *
from dev0s.classes.response import ResponseObject
from dev0s.classes.response import response as _response_
from dev0s.classes.defaults import objects
from dev0s.classes.defaults.defaults import defaults

# pip imports.
import pexpect
import signal as _signal_

# kill pids.
def kill(
	# option 1:
	# the process id.
	pid=None, 
	# option 2:
	# all processes that includes.
	includes=None,
	# root permission required.
	sudo=False,
	# loader.
	log_level=0,
):
	# kill includes.
	loader = None
	if includes != None:
		response = processes(includes=includes, sudo=sudo)
		if not response.success: return response
		c = 0
		for pid, info in response.processes.items():
			response = kill(sudo=sudo, pid=pid, loader=loader)
			if not response.success: return response
			c += 1
		if c == 0:
			return _response_.error(f"No processes found.")
		elif c == 1:
			return _response_.success(f"Successfully killed {c} process.")
		else:
			return _response_.success(f"Successfully killed {c} processes.")

	# kill pid.
	else:
		if log_level >= 1:
			loader = console.Loader(f"Killing process {pid}.")
		_sudo_ = Boolean(sudo).string(true="sudo ", false="")
		output = utils.__execute_script__(f"{_sudo_}kill {pid}")
		if output in ["terminated",""]:
			if output in [""]:
				response = processes(includes=pid)
				if not response.success: response
				try: 
					response.processes[pid]
					if log_level >= 1: loader.stop(success=False)
					return _response_.error(f"Failed to stop process {pid}.", log_level=log_level)
				except KeyError: a=1
			if log_level >= 1: loader.stop()
			return _response_.success(f"Successfully killed process {pid}.")
		else:
			if log_level >= 1: loader.stop(success=False)
			return _response_.error(f"Failed to stop process {pid}, error: {output}", log_level=log_level)

# list all processes.
def processes(
	# root permission.
	sudo=False,
	# all processes that include a str.
	includes=None,
	# banned process names.
	banned=["grep"],
):
	_sudo_ = Boolean(sudo).string(true="sudo ", false="")
	if isinstance(includes, str):
		command = f"""{_sudo_}ps -ax | grep "{includes}" | """
	else:
		command = f"""{_sudo_}ps -ax | """
	#output = utils.__execute_script__(command + """awk '{print $1"|"$2"|"$3"|"$4"}' """)
	output = utils.__execute_script__(command + """awk '{$3="";print $0}' """)
	processes = {}
	for line in output.split("\n"):
		if line not in ["", " "]:
			array = line.split(" ")
			pid = array.pop(0)
			tty = array.pop(0)
			for i in array:
				process = array[0]
				if process not in ["", " "]: break
				else: array.pop(0)
			command = Array(array).string(joiner=" ")
			#try:
			#	pid,tty,_,process,command = line.split("|")
			#except ValueError: raise ValueError(f"Unable to unpack process line: [{line}], expected format: [4] seperator: [|].")
			if process not in banned:
				processes[pid] = {
					"pid":pid,
					"tty":tty,
					"process":process,
					"command":command,
				}
	return _response_.success(f"Successfully listed {len(processes)} processes.", {
		"processes":processes,
	})

# execute.
def execute(
	# Notes:
	#   returns a dev0s.code.OutputObject object (very similair to ResponseObject).
	#
	# Mode:
	#   option 1:
	#     the command in str format, the command is saved to a script & then executed).
	command="ls .",
	#     joiner for when command is in list format.
	joiner=" ",
	#   option 2: the path to script.
	path=None,
	#
	# Executement:
	#   the executable.
	executable="sh",
	#   the arguments passed to the (saved) script.
	arguments=[], 
	#
	# Options:
	#   asynchronous process.
	async_=False,
	#	await asynchronous child (sync process always awaits).
	wait=False,
	#	kill process when finished (async that is not awaited is never killed).
	kill=True,
	#   the subprocess shell parameter.
	shell=False,
	#   serialize output to dict (expect literal dictionary / json output).
	serialize=False,
	#
	# Input (sync only):
	#   send input to the command.
	#	  undefined: send no input & automatically await the process since input is always sync.
	#	  dict instance: selects "and" mode ; send expected inputs and their value & return error when one of them is missing.
	#	  list[dict] instance: send all dictionaries in the list (default dict behaviour so one of the keys in each dict is expected).
	input=None,
	#   the input timeout (float) (list with floats by index from input)
	timeout=1.0,
	#   do not throw an error when the input is missing or not expected when optional is disabled (bool).
	optional=False, 
	#
	# Logging.
	# the loader (str, Loader).
	loader=None,
	# stop the loader at the end of the request (bool).
	stop_loader=True,
	#   the log level.
	log_level=defaults.options.log_level,
	#
	# System functions.
	#   add additional attributes to the spawn object.
	__spawn_attributes__={},
	#
):	

	# checks,
	if input != None and not isinstance(input, (dict, Dictionary, list, Array)): 
		raise Exceptions.InvalidUsage(f"<dev0s.code.execute>: Parameter [input] requires to be be a [dict, Dictionary, list, Array], not [{iput.__class__.__name__}].")

	# loader.
	if isinstance(loader, (str,String)) and 2 > log_level >= 0:
		loader = console.Loader(loader)

	# vars.
	delete = False
	if path == None:
		delete = True
		path = f"/tmp/tmp_script_{String('').generate()}"
		if isinstance(command, list): command = Array(array=command).string(joiner=joiner)
		Files.save(path, command)
		response_str = f"command ({command})"
	else:
		response_str = f"script ({path})"

	# execute with input.
	#if isinstance(input, (dict, Dictionary, list, Array)):

	# checks.
	#if async_: 
	#	raise Exceptions.InvalidUsage(f"<dev0s.code.execute>: Parameters [input] & [async] are not compatible, select either one.")

	# spawn.
	l = []
	for i in arguments: l.append(f'"{i}"')
	arguments = Array(l).string(joiner=' ')
	if log_level >= 8: print(f"Spawn: [ $ {executable} {path} {arguments}]",)
	spawn = Spawn(
		command=f"{executable} {path} {arguments}",
		#async_=async_, # does not work.
		response_str=response_str,
		log_level=log_level,
		attributes=__spawn_attributes__,
	)
	spawn.echo = False
	if isinstance(timeout, (int,float,Integer)):
		spawn.timeout = int(timeout)
	
	# start.
	response = spawn.start()
	if not response.success: 
		if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
		return OutputObject(error=f"Failed to start {response_str}, error: {response.error}", log_level=log_level)
	
	# check crashed
	response = spawn.crashed()
	if not response.success:  
		if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
		return OutputObject(error=response.error, log_level=log_level)

	# has already exited.
	elif spawn.exit_status in [0]:
		
		# get output.
		response = spawn.wait(timeout=1.0)
		if not response.success: 
			if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
			return OutputObject(error=response.error, log_level=log_level) # exit status.
		output = response.output

		#

	# proceed if already finished.
	elif spawn.exit_status not in [0]:

		# await.
		#success = False
		#for i in range(10):
		#	if spawn.running: 
		#		success = True
		#		break
		#	time.sleep(1)
		#if not success: 
		#	if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
		#	return OutputObject(error=f"Unable to start {response_str}.", log_level=log_level)

		# str input.
		if isinstance(input, (list, Array)):
			if len(input) > 0:
				str_input = Array(list(input[0].keys())).string(joiner=", ")
			else:
				str_input = Array(input).string(joiner=", ")
		elif isinstance(input, (dict, Dictionary)):
			str_input = Array(list(input.keys())).string(joiner=", ")
		else: 
			str_input = f"[{input}]"

		# send optional input.
		error_end_of_file = None
		if not async_ and isinstance(input, (list, Array, dict, Dictionary)) and len(input) > 0:

			# expect one of the keys in the dictionary.
			def process_dict_input(dictionary):

				# vars.
				expect = list(dictionary.keys())
				send = list(dictionary.values())

				# expect .
				if log_level >= 8:
					print(f"Expecting one of the following inputs: {expect}.")
				response = spawn.expect(expect=expect, send=send, timeout=timeout)
				if not response.success:
					if "None of the specified inputs were expected." in response.error:
						if optional:
							return _response_.error(f"Unable to find the expected input but still success since it is optional.")
						else:
							return _response_.error(f"Specified input [{Array(expect).string(joiner=', ')}] was not expected.")
					else: return response
				if log_level >= 8:
					print("Send response message:",response.message)

				# success.
				return _response_.success("Success.")

			""" check expecting. """
			expecting = True
			if not spawn.expecting:
				expecting = False
				if not optional:
					if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
					if log_level >= 1:
						return OutputObject(error=f"Failed to send expected input {str_input} to {response_str}, child is not expecting any input [{spawn.child}].", log_level=log_level)
					else:
						return OutputObject(error=f"Failed to send expected input {str_input} to {response_str}, child is not expecting any input.", log_level=log_level)

			# limit not expecting by optional.
			if expecting:

				# send all dicts in the list (list instance).
				error_end_of_file = False
				if isinstance(input, (list, Array)):
					for _input_ in input:
						response = process_dict_input(_input_)
						if not response.success: 
							if "End of file" in response.error:
								error_end_of_file = True
							else: 
								# str input.
								if isinstance(_input_, (list, Array)):
									str_input = Array(_input_).string(joiner=", ")
								elif isinstance(_input_, (dict, Dictionary)):
									str_input = Array(list(_input_.keys())).string(joiner=", ")
								else: 
									str_input = f"[{_input_}]"
								if optional:
									break
								else:
									if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
									return OutputObject(error=f"Failed to send one of the expected input(s) {str_input}, error: {response.error}", log_level=log_level)

				# send one of the keys (dict instance).
				elif isinstance(input, (dict, Dictionary)):
					response = process_dict_input(input)
					if not response.success: 
						if "End of file" in response.error:
							error_end_of_file = True
						elif not optional:
							if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
							return OutputObject(error=f"Failed to send one of the expected input(s) {str_input}, error: {response.error}", log_level=log_level)

				""" check no input left (does not work properly).
				if not error_end_of_file and spawn.expecting:
					try: after = spawn.child.after.decode()
					except : after = spawn.child.after
					if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
					return OutputObject(error=f"Failed to execute {response_str}, still expecting: [{after}].", log_level=log_level)
				"""		

		# do not get or kill when async.
		output = None
		if not async_:

			# check crashed.
			response = spawn.crashed()
			if not response.success:  
				if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
				return OutputObject(error=response.error, log_level=log_level)

			# always await sync.
			response = spawn.wait()
			if not response.success: 
				if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
				return OutputObject(error=response.error, log_level=log_level)
			output = response.output
			if error_end_of_file != None and error_end_of_file: 
				if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
				if log_level >= 1:
					return OutputObject(error=f"Failed to send expected input {str_input} to {response_str} (#234343) (output: {output}) (child: {spawn.child}).", log_level=log_level)
				else:
					return OutputObject(error=f"Failed to send expected input {str_input} to {response_str} (#234343) (output: {output}).", log_level=log_level)

			# check kill.
			if kill and spawn.running:
				if log_level >= 8: print(f"Killing process {response_str}.")
				response = spawn.kill()
				if not response.success: 
					if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
					return OutputObject(error=f"Failed to kill {response_str}, error: {response.error}", log_level=log_level)

		# async.
		elif async_:

			# check exit status.
			response = spawn.read(wait=False)
			if not response.success: 
				if not response.success: 
					if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
					return OutputObject(error=f"Failed to retrieve output from spawn {response_str}, error: {response.error}", log_level=log_level)
			if spawn.child.exitstatus not in [0, None]:
				if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
				return OutputObject(error=f"{response_str} returned exit status: [{spawn.child.exitstatus}] (output: {self.read(wait=False, __safe__=True).output}).", log_level=log_level)

			# await async.
			if wait:

				# await.
				response = spawn.wait()
				if not response.success: 
					if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
					return OutputObject(error=response.error, log_level=log_level) # exit status.
				output = response.output

				if spawn.child.exitstatus not in [0, None]:
					if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
					return OutputObject(error=f"{response_str} returned exit status: [{spawn.child.exitstatus}] (output: {self.read(wait=False, __safe__=True).output}).", log_level=log_level)
				
				# check kill.
				if kill and spawn.running:
					if log_level >= 8: print(f"Killing process {response_str}.")
					response = spawn.kill()
					if not response.success: 
						if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
						return OutputObject(error=f"Failed to kill {response_str}, error: {response.error}", log_level=log_level)

	# handler.
	if delete: Files.delete(path)
	if serialize:
		try: response = _response_.ResponseObject(output)
		except Exception as e: 
			if loader != None: loader.stop(success=False)
			if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
			return OutputObject(error=f"Failed to serialize (output: {output}).", log_level=log_level)
		if not response.success: 
			if stop_loader and isinstance(loader, console.Loader): loader.stop(success=False)
			return OutputObject(error=f"Encoutered an error in the serialized response, error: {response.error}", log_level=log_level)
	if stop_loader and isinstance(loader, console.Loader): loader.stop()
	return OutputObject(message=f"Succesfully executed {response_str}.", log_level=log_level, attributes={
		"output":output,
		"process":spawn,
		"pid":spawn.child.pid,
		"running":spawn.running,
		"exit_status":spawn.exit_status,
	})


	"""
	# excute default without input.
	else:

		# excute.
		if async_:
			proc = subprocess.Popen(
				[executable, path] + arguments,
				shell=shell,)
			return OutputObject(message=f"Succesfully executed {response_str}.", log_level=log_level, attributes={
				"output":"",
				"process":proc,
				"pid":proc.pid,
			})
		try:
			proc = subprocess.run(
				[executable, path] + arguments,
				check=True,
				capture_output=True,
				text=True,
				shell=shell,
			)
		except subprocess.CalledProcessError as error:
			exception = String(str(error))
			if "Command '[" in exception:
				to_replace = f"'{exception.slice_array()}'"
				exception = exception.replace(str(to_replace), "[ ... ]")
			error_, output = error.stderr, error.output
			if isinstance(error_, bytes): error_ = error_.decode()
			if isinstance(output, bytes): output = output.decode()
			if log_level <= 0:
				return OutputObject(error=f"Failed to execute {response_str}, (exception: {exception}), (error: {error_}).", log_level=log_level)
			else:
				return OutputObject(error=f"Failed to execute {response_str}, (exception: {exception}), (output: {output}), (error: {error_}).", log_level=log_level)
		error_, output = proc.stderr, proc.stdout
		if isinstance(error_, bytes): error_ = error_.decode()
		if isinstance(output, bytes): output = output.decode()
		if error_ != "":
			if log_level <= 0:
				return OutputObject(error=f"Failed to execute {response_str}, (error: {error_}).", log_level=log_level)
			else:
				return OutputObject(error=f"Failed to execute {response_str}, (output: {output}), (error: {error_}).", log_level=log_level)
		if len(output) > 0 and output[len(output)-1] == "\n": output = output[:-1]
		if serialize:
			try: response = _response_.ResponseObject(output)
			except Exception as e: 
				if loader != None: loader.stop(success=False)
				return OutputObject(error=f"Failed to serialize (output: {output}).", log_level=log_level)
			return response
		else:
			return OutputObject(message=f"Succesfully executed {response_str}.", log_level=log_level, attributes={
				"output":output,
				"process":proc,
				"pid":None,
			})
	"""

# the spawn object class (for inputs).
class Spawn(objects.Object):
	def __init__(self,
		#
		# Should be initialized with function: dev0s.code.execute
		#
		# the full command (str) (#1).
		command="ls",
		# asynchronous.
		async_=False,
		# the log level.
		log_level=defaults.options.log_level,
		# additional attributes.
		attributes={},
		# system options.
		response_str=None,
	):

		# defaults.
		objects.Object.__init__(self, traceback="dev0s.code.execute.spawn")
		self.assign(attributes)

		# args.
		self.command = command
		self.async_ = async_
		self.log_level = log_level
		self.response_str = response_str

		# vars.
		self.child = None
		self.__secrets__ = []

		#

	# start the process.
	def start(self):
		try:
			self.child = pexpect.spawn(self.command)
		except Exception as e:
			return _response_.error(f"Failed to spawn {self.response_str}, error: {e}.")
		# check exit status.
		if self.child.exitstatus not in [0, None]:
			return OutputObject(error=f"{self.response_str} returned exit status: [{spawn.child.exitstatus}] (output: {self.read(wait=False, __safe__=True).output}).")
		self.__output__ = ""
		return _response_.success(f"Successfully spawned {self.response_str}.")
		#

	# check expect input and optionally send input.
	def expect(self,
		# the expected data parameter (#1).
		#	str instantce: expect a single identifier.
		#	list instance: expect one of the provided identifiers & return the found one if success.
		expect=["Password*"],
		# the optional data to send (#2).
		#	none instance: do not send anything.
		#	str instance: the data to send.
		#	list/tuple instance: send value of index from expected expect (required expect to be a list, Array & the indexes of [expect, send] be match).
		send=None,
		# the timeout (float).
		timeout=1.0,
	):
		# check none timeout.
		if timeout == None: timeout = 1.0

		# single expect.
		if isinstance(expect, str):
			if isinstance(send, (list, Array)): 
				raise Exceptions.InvalidUsage(f"{self.__traceback__(function='expect', parameter='send')}: parameter [send] requires to be be a [str, String] when parameter [expect] is a [{expect.__class__.__name__}], not [{send.__class__.__name__}].")

			""" add to lines to output (before) (adds nothing when empty) (do not use self.expecting before adding the output). """
			response = self.read(wait=False)
			if not response.success: return response

			""" check expect & send if provided. """
			c = 1
			try:
				#if self.async_:
				#	g = yield from self.child.expect(expect, timeout=timeout, async_=True)
				#	r = next(g)
				#else:
				r = self.child.expect(expect, timeout=timeout)
				if not isinstance(r, (int, Integer, float)):
					return _response_.error(f"Expected [{expect}] is not the child's expected input (#873465).")
				c = 2
				if self.log_level >= 8: print(f"Found expected: {expect}")
				if send != None:
					if self.log_level >= 8: print(f"Attempting to send: {send}")
					#if self.async_:
					#	g = yield from self.child.sendline(str(send), async_=True)
					#	r = next(g)
					#else:
					r = self.child.sendline(str(send))
					self.__secrets__.append(str(send))
					if self.log_level >= 8: print(f"Succesfully sended: {send}")
			except pexpect.exceptions.TIMEOUT:
				if c == 1:
					return _response_.error(f"Expected [{expect}] is not the child's expected input, error: Timeout (during epxecting).")
				else:
					return _response_.error(f"Failed to send expected [{expect}], error: Timeout (during sending input).")
			except pexpect.exceptions.EOF:
				if c == 1:
					return _response_.error(f"Expected [{expect}] is not the child's expected input, error: End of file (during epxecting).")
				else:
					return _response_.error(f"Failed to send expected [{expect}], error: End of file (during sending input).")
			except Exception as e:
				if c == 1:
					return _response_.error(f"Expected [{expect}] is not the child's expected input, error: {e}.")
				else:
					return _response_.error(f"Failed to send expected [{expect}], error: {e}.")

			
			""" add to lines to output (after) (adds nothing when empty) (do not use self.expecting before adding the output). """
			response = self.read(wait=False)
			if not response.success: return response

			# handler.
			if send != None:
				return _response_.success(f"Successfully send expected input ({expect}).")
			else:
				return _response_.success(f"Successfully retrieved expected input ({expect}).")

		# list expect.
		elif isinstance(expect, (list, Array)):
			index = 0
			for _expect_ in expect:
				if isinstance(send, (list,Array)):
					try:
						_send_ = str(send[index])
					except:
						raise Exceptions.InvalidUsage(f"{self.__traceback__(function='expect', parameter='send')}: parameter [send] and parameter [expect] do not have the same indexes.")
				else:
					_send_ = str(send)
				if self.log_level >= 8: print(f"Checking optional expect: {_expect_}")
				response = self.expect(expect=_expect_, timeout=timeout, send=_send_)
				if not response.success and "is not the child's expected input" not in response.error: return response
				elif response.success:
					return _response_.success(f"Successfully {Boolean(send).string(true='send', false='retrieved')} the expected input(s).", {
						"expected":_expect_,
						"index":index,
					})
				index += 1
			return _response_.error(f"None of the specified input was expected.")

		# invalid usage.
		else:
			raise Exceptions.InvalidUsage(f"{self.__traceback__(function='expect', parameter='expect')}: parameter [expect] requires to be be a [Dictionary], not [{config.__class__.__name__}].")

	# read output.
	def read(self, 
		# with await False it reads only the printed output regardless the status & never throws timeout.
		wait=False, 
		# the timeout, leave None for no timeout.
		timeout=None,
		# the live boolean (bool) (prints live logs to console when enabled) (leave None to use self.log_level >= 1).
		live=None, 
		# system variables.
		#   safe True always a response.output variable upon error the response.output is "".
		__safe__=False,
	):

		# clean line.
		def clean_line(line):
			if isinstance(line, bytes): line = line.decode()
			if line in [" password:"]: return None
			for secret in self.__secrets__:
				for strip_last in [secret]:
					if secret not in ["", " ", "  ", "   ", "    "] and len(line) >= len(strip_last) and line[-len(strip_last):] == strip_last: 
						return line[:-len(strip_last)]
			return line
		def clean_output(output):
			_output_, b = "", output.replace("\r","").replace("\t","").split("\n")
			c, m = 0, len(b)-1
			for line in b: 
				l = clean_line(line)
				if l not in [None]:
					if c < m: _output_ += l+"\n"
					else: _output_ += l
				c += 1
			return _output_
		def log_output(output):
			while True:
				if len(output) >= 1 and output[len(output)-1] == "\n": output = output[:-1]
				else: break	
			print(output)

		# checks.
		if live == None: live = self.log_level >= 1

		# old timeout.
		old = self.child.timeout

		# version 2: reads only the printed lines and stops when the timeout is hit so there is no timeout required. 
		output_logs_allowed = True
		if not wait:
			self.child.timeout = 0.25
			output_logs_allowed = False
			output, c = "", 0
			for line in range(1000000000):
				try:
					#if self.async_:
					#	g = yield from self.child.readline(line, async_=True)
					#	output += next(g).decode()
					#else:
					new = self.child.readline(line).decode()
					if new == "": 
						c += 1
						if c >= 25: break
					else: 
						output_logs_allowed = True
						c = 0
						output += new
				except pexpect.exceptions.TIMEOUT:
					break
				except pexpect.exceptions.EOF:
					break
				except Exception as e:
					if timeout != None: self.child.timeout = old
					e = str(e)
					if self.log_level <= 0: e = str(e).split("\n")[0] # strip the full child from the pexpect error message.
					while True:
						if len(e) >= 1 and e[len(e)-1] == ".": e = e[:-1]
						else: break
					if "Timeout" in str(e):
						response = _response_.error(f"{e} (most likely the command is still expecting input).")
						if __safe__: response.output = ""
						return response
					else:
						response = _response_.error(f"{e}.")	
						if __safe__: response.output = ""
						return response

		# version 1: throws an error when unable to read output if timeout is defined, so it can be used to detect if input is expected. 
		else:

			# handle output.
			self.child.timeout = timeout
			try:
				#output = self.child.read().decode()
				output = ""
				#if self.async_:
				#	g = yield from self.child.readlines(async_=True)
				#	lines = next(x)
				#else:
				lines = self.child.readlines()
				if lines == []:
					output_logs_allowed = False
				else:
					for i in lines: output += i.decode()
			except Exception as e:
				self.child.timeout = old
				e = str(e)
				if self.log_level <= 0: e = str(e).split("\n")[0] # strip the full child from the pexpect error message.
				while True:
					if len(e) >= 1 and e[len(e)-1] == ".": e = e[:-1]
					else: break
				if "Timeout" in str(e):
					response = _response_.error(f"{e} (most likely the command is still expecting input).")
					if __safe__: response.output = ""
					return response
				else:
					response = _response_.error(f"{e}.")
					if __safe__: response.output = ""
					return response

		# clean output.
		output = clean_output(output)
		if output_logs_allowed and live: log_output(output)

		# handler.
		self.child.timeout = old
		self.__output__ += output
		self.__output__ = clean_output(self.__output__)
		return _response_.success("Successfully retrieved the child's output.", {
			"output":str(self.__output__),
			"new_output":output,
		})


		#

	# kill the process.
	def kill(self):
		
		# handle output.
		"""
		killed = None
		try:
			#if self.async_:
			#	g = yield from self.child.terminate(force=True, async_=True)
			#	killed = next(x)
			#else:
			killed = self.child.terminate(force=True)
		except Exception as e:
			return _response_.error(f"Failed to kill process [{self.command}], error: {e}.")
		if killed == None:
			return _response_.error(f"Failed to kill process [{self.command}] (#452983).")
		if not killed:
		"""
		
		# handle output.
		response = kill(pid=self.pid, log_level=self.log_level)
		if not response.success:
			return _response_.error(f"Unable to kill process [{self.command}].")
		else:
			return _response_.success(f"Successfully killed process [{self.command}].")

		#

	# wait for the process to finish.
	def wait(self, 
		# the live boolean (bool) (prints live logs to console when enabled) (leave None to use self.log_level >= 1).
		live=None, 
		sleeptime=1,
		# the timeout (leave None to ignore).
		timeout=None,
	):
		
		# checks.
		if live == None: live = self.log_level >= 1

		# live
		if live:
			while self.running:
				response = self.read(wait=False, live=live)
				if not response.success: return response
				if self.running:
					time.sleep(sleeptime)
			return self.wait(live=False, timeout=timeout)

		# not live.
		else:

			# read & wait with no timeout
			response = self.read(wait=True, timeout=timeout)
			if response.success:
				while True:
					if len(response.output) > 0 and response.output[len(response.output)-1] == "\n": response.output = response.output[:-1]
					else: break
			return response

		#

	# check if process is crashed.
	def crashed(self):
		response = self.read(wait=False)
		if not response.success:  return response
		if self.exit_status not in [0, None]:
			return _response_.error(f"{self.response_str} returned exit status: [{self.exit_status}] (output: {self.read(wait=False, __safe__=True).output}).")
		return _response_.success(f"{self.response_str} is not crashed.")

	# is expecting input.
	@property
	def expecting(self):

		# version 2.
		if self.exit_status not in [None]: return False
		elif self.child.closed: return False
		elif not self.running: return False
		else:


			""" add to lines to output (adds nothing when empty) (to prevent the output reset called the proceeding code) . """
			response = self.read(wait=False)
			if not response.success: return response

			""" output throws timeout error when still expecting & when with success a process has finiished. """
			response = self.read(wait=True, timeout=1.0) 
			if not response.success:
				if "still expecting input" in response.error: return True
				else: raise ValueError(f"Unknown behaviour while detecting @spawn.expecting, error: {response.error}")
			else: return False

			#
		raise ValueError(f"Unable to determine is spawn is expecting, child: {self.child}.")

		""" version 1.
		try:
			# renaming "" to pexpect.EOF removes the output.
			self.child.expect("", timeout=1)
			return False
		except pexpect.exceptions.EOF:
			return True
		except:
			return False
		raise ValueError(f"Unable to determine is spawn is expecting, child: {self.child}.")
		"""

	# is running.
	@property
	def running(self):
		"""if self.async_:
			boolean = yield from self.child.isalive(async_=True)
		else:
			boolean = self.child.isalive()
		"""
		boolean = self.child.isalive()
		return boolean
		#
	# exit status.
	@property
	def exit_status(self):
		"""if self.async_:
			boolean = yield from self.child.exitstatus
		else:
			boolean = self.child.exitstatus
		"""
		boolean = self.child.exitstatus
		return boolean
		#
	# output property.
	@property
	def output(self):
		return self.__output__
	# pid.
	@property
	def pid(self):
		return self.child.pid
	

	#

# the output object class (very similair to the ResponseObject).
class OutputObject(ResponseObject):
	def __init__(self, 
		#
		# The return object from function: dec0s.code.execute
		# The OutputObject object is very similair to the ResponseObject.
		#
		# the success message (param #1).
		message=None,
		# the attributes (param #2).
		attributes={},
		# the error message (param #3).
		error=None,
		# the log level.
		log_level=defaults.options.log_level,
	):
		dictionary = {
			"success":error == None,
			"message":message,
			"error":error,
			"log_level":log_level,
		}
		for key,value in attributes.items():
			dictionary[key] = value
		ResponseObject.__init__(self, dictionary)
		if self.success:
			try: 
				if self.output == None: raise AttributeError("")
			except AttributeError: self.output = ""
		else: self.output = str(self.error) # keep this way for system functions.
		self.lines = self.output.split("\n")

	# representations.
	def __len__(self):
		return len(self.output)
		#
	def __bool__(self):
		return bool(self.success)
		#

		#

	# support '==' & '!=' operator.
	def __eq__(self, dictionary):
		if dictionary.__class__.__name__ in ["str", "String", "NoneType", "int", "float", "Integer"]:
			return str(self.output) != str(dictionary)
		elif isinstance(dictionary, dict):
			return str(self.sort()) == str(Dictionary(dictionary).sort())
		elif isinstance(dictionary, Dictionary):
			return str(self.sort()) == str(dictionary.sort())
		else:
			try:
				return str(self.sort()) == str(dictionary.sort())
			except:
				return False
	def __ne__(self, dictionary):
		if dictionary.__class__.__name__ in ["str", "String", "NoneType", "int", "float", "Integer"]:
			return str(self.output) != str(dictionary)
		elif isinstance(dictionary, dict):
			return str(self.sort()) != str(Dictionary(dictionary).sort())
		elif isinstance(dictionary, Dictionary):
			return str(self.sort()) != str(dictionary.sort())
		else:
			try:
				return str(self.sort()) != str(dictionary.sort())
			except:
				return False

	# support 'in' operator.
	def __contains__(self, string):
		if isinstance(string, (list, Array)):
			for i in string:
				if str(i) in str(self.output):
					return True
			return False
		else:
			return str(string) in str(self.output)
		#

	# object instance.
	def instance(self):
		return "OutputObject"
		#
	# assign to ResponseObject.
	def response(self):
		return _response_.response(self.dict())
		#

	#


#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.defaults.exceptions import Exceptions
from dev0s.classes.defaults.files import *
from dev0s.classes.response import ResponseObject
from dev0s.classes.response import response as _response_

# defaults.options.
JSON = "-j" in sys.argv or "--json" in sys.argv

# traceback.
class Traceback(object):
	def __init__(self,
		# the imported traceback (#1).
		traceback="Traceback",
		# the raw traceback (#2).
		raw_traceback="Object",
	):

		# defaults.
		self._traceback_ = traceback
		self._raw_traceback_ = traceback

		#
	
	# imported traceback.
	@property
	def traceback(self):
		return self.__traceback__()
	def __traceback__(self, 
		# the function name (optional) (#1).
		function=None,
		# the parameter name (optional) (#2).
		parameter=None,
		# the attrbiute name (optional) (#3).
		attribute=None,
		# raw traceback.
		raw=False,
	):
		options = ""
		if attribute != None:
			options += f".{attribute}"
		elif function != None:
			options += f".{function}"
			if parameter != None:
				options += f".{parameter}"
		if raw:
			return f"<{self._raw_traceback_}{options}>"
		else:
			return f"<{self._traceback_}{options}>"

	#

# object.
class Object(Traceback):
	def __init__(self, 
		# attributes (dict) (#1)
		attributes={},
		# the imported traceback.
		traceback="Object",
		# the raw traceback.
		raw_traceback="Object",
	):

		# defaults.
		Traceback.__init__(self, traceback=traceback, raw_traceback=raw_traceback)
		self.assign(attributes)
		#

	# iterations.
	def items(self,
		# the keys to get (leave default to unpack all keys).
		#	list instance: checks if the key is present if not it throws an error when [safe] is disabled
		#	dict instance: automatically enables [safe] and returns the key's value as default when missing.
		keys=["*"],
		# with safe disabled it throws an error when one of the specified keys does not exist.
		safe=True,
	):
		
		# safe dict keys.
		selected = {}
		if isinstance(keys, (dict, Dictionary)):
			for key,value in keys.items():
				skip = False
				try:
					dict_value = self[key]
				except:
					dict_value = value
					skip = False
				if not skip:
					selected[key] = dict_value
			return selected.items()

		# array keys.
		elif isinstance(keys, (list, Array)) and "*" not in keys:
			for key in keys:
				skip = False
				if safe:
					try:
						dict_value = self[key]
					except:
						skip = True
				else:
					dict_value = self[key]
				if not skip:
					selected[key] = dict_value
			return selected.items()
		
		# all keys from dict.
		else:
			return vars(self).items()

		#
	def keys(self, 
		# the keys to get (leave default to unpack all keys).
		#	list instance: checks if the key is present if not it throws an error when [safe] is disabled
		#	dict instance: automatically enables [safe] and returns the key's value as default when missing.
		keys=["*"],
		# with safe disabled it throws an error when one of the specified keys does not exist.
		safe=True,
	):
		
		# safe dict keys.
		selected = {}
		if isinstance(keys, (dict, Dictionary)):
			for key,value in keys.items():
				skip = False
				try:
					dict_value = self[key]
				except:
					dict_value = value
					skip = False
				if not skip:
					selected[key] = dict_value
			return list(selected.keys())

		# array keys.
		elif isinstance(keys, (list, Array)) and "*" not in keys:
			for key in keys:
				skip = False
				if safe:
					try:
						dict_value = self[key]
					except:
						skip = True
				else:
					dict_value = self[key]
				if not skip:
					selected[key] = dict_value
			return list(selected.keys())
		
		# all keys from dict.
		else:
			return list(vars(self).keys())

		#
	def values(self):
		return list(vars(self).values())
	

	# assign self variables by dictionary.
	def assign(self, 
		# the dictionary to self assign.
		dictionary, 
		# serialize dictionary from str to object.
		serialize=True,
		# the keys to get from the dict (leave default to unpack the present keys).
		#	list instance: checks if the key is present if not it throws an error when [safe] is disabled
		#	dict instance: automatically enables [safe] and returns the key's value as default when missing.
		keys=["*"],
		# with safe disabled it throws an error when one of the specified keys does not exist.
		safe=True,
	):
		if not isinstance(dictionary, (dict,Dictionary)):
			raise TypeError("You can only self assign with a dictionary as parameter.")
		if not isinstance(keys, (dict, Dictionary, list, Array) ):
			raise TypeError(f"Parameter [keys] requiers to be a [dict, Dictionary, list, Array] not [{keys.__class__.__name__}].")
		# safe dict keys.
		if isinstance(keys, (dict, Dictionary)):
			safe = True
			for key,value in keys.items():
				skip = False
				if safe:
					try:
						dict_value = dictionary[key]
					except:
						dict_value = value
						skip = False
				else:
					dict_value = dictionary[key]
				if not skip:
					if serialize:
						if dict_value in ["True", "true", "TRUE"]: dict_value = True
						elif dict_value in ["False", "false", "FALSE"]: dict_value = False
						elif dict_value in ["None", "none", "null"]: dict_value = None
					self[key] = dict_value
		# array keys.
		elif isinstance(keys, (list, Array)) and "*" not in keys:
			for key in keys:
				skip = False
				if safe:
					try:
						dict_value = dictionary[key]
					except:
						skip = True
				else:
					dict_value = dictionary[key]
				if not skip:
					if serialize:
						if dict_value in ["True", "true", "TRUE"]: dict_value = True
						elif dict_value in ["False", "false", "FALSE"]: dict_value = False
						elif dict_value in ["None", "none", "null"]: dict_value = None
					self[key] = dict_value
		# all keys from dict.
		else:
			for key,value in dictionary.items():
				if serialize:
					if value in ["True", "true", "TRUE"]: value = True
					elif value in ["False", "false", "FALSE"]: value = False
					elif value in ["None", "none", "null"]: value = None
				self[key] = value

	# get all attributes.
	def attributes(self, 
		# the keys to get (leave default to unpack all keys).
		#	list instance: checks if the key is present if not it throws an error when [safe] is disabled
		#	dict instance: automatically enables [safe] and returns the key's value as default when missing.
		keys=["*"],
		# with safe disabled it throws an error when one of the specified keys does not exist.
		safe=True,
	):
		return self.dict(keys=keys, safe=safe)
	def dict(self, 
		# the keys to get (leave default to unpack all keys).
		#	list instance: checks if the key is present if not it throws an error when [safe] is disabled
		#	dict instance: automatically enables [safe] and returns the key's value as default when missing.
		keys=["*"],
		# with safe disabled it throws an error when one of the specified keys does not exist.
		safe=True,
	):
		dictionary = {}
		for key, value in self.items(keys=keys, safe=safe):
			dictionary[key] = value
		return dictionary

		#

	# unpack attributes.
	def unpack(self, 
		# the key / keys / defaults parameter (#1).
		# str instance:
		#   unpack the str key
		# list instance:
		#   unpack all keys in the list.
		# dict instance:
		#   unpack all keys from the dict & when not present return the key's value as default.
		keys,
	):
		defaults_ = {}
		if isinstance(keys, (dict, Files.Dictionary)):
			if isinstance(keys, dict):
				defaults_ = dict(keys)
				keys = list(keys.keys())
			else:
				defaults_ = keys.dict()
				keys = keys.keys()
		elif isinstance(keys, str):
			keys = [keys]
		unpacked = []
		for key in keys:
			value, set = None, True
			try: 
				value = self[key]
			except KeyError: 
				try:
					value = defaults_[key]
				except KeyError: 
					set = False
			if not set:
				raise Exceptions.UnpackError(f"Dictionary does not contain attribute [{key}].")
			unpacked.append(value)
		if len(unpacked) == 1:
			return unpacked[0]
		else:
			return unpacked

	# support item assignment.
	def __setitem__(self, key, value):
		setattr(self, key, value)
	def __getitem__(self, key):
		return getattr(self, key)
	def __delitem__(self, key):
		delattr(self, key)

		#


	#

# thread.
class Thread(Object, threading.Thread):
	def __init__(self, 
		# the threads id (#1).
		id="Thread",
		# the imported traceback.
		traceback="Thread",
		# the raw traceback.
		raw_traceback="Thread",
		# the threads log level.
		log_level=-1,
	):

		# defaults.
		Object.__init__(self, traceback=traceback, raw_traceback=raw_traceback)
		threading.Thread.__init__(self)

		# args.
		self.log_level = log_level

		# vars.
		try: self.id = id
		except: a=1
		self.__status__ = "initialized"
		self.__response__ = None
		
		#

	# default start (should not be overwritten, use: __run__.
	def run(self):
		self.log(f"Starting thread {self.id}.")
		self.__status__ = "running"
		response = None
		try:
			response = self.__run__()
		except AttributeError as e:  
			if "has no attribute '" in str(e):
				missing = str(e).split("has no attribute '")[1].split("'")[0]
				if missing == "__run__": raise Exceptions.InvalidUsage(f"Thread [{self.id}] is missing the main [__run__] function.")
				else: raise AttributeError(str(e))
			else: raise AttributeError(str(e))
		if isinstance(response, ResponseObject):
			if response.success:
				self.send_crash(response=response)
			else:
				self.send_stop(response=response)
		else:
			self.send_stop(response=response)
		return self.send_stop(f"Thread [{self.id}] has finished running.")

	# safely start & stop the thread.
	def safe_start(self, timeout=120, sleeptime=1):
		self.log(f"Stopping thread {self.id}.")
		self.start()
		for i in range(int(timeout/sleeptime)):
			if self.running: break
			time.sleep(sleeptime)
		if not self.running: 
			return _response_.error(f"Unable to start thread {self}.", log_level=self.log_level, required_log_level=0)
		return _response_.success(f"Successfully started thread {self.id}.", log_level=self.log_level, required_log_level=0)
	def safe_stop(self, timeout=120, sleeptime=1):
		
		# check already crashed.
		if self.crashed:
			return _response_.error(f"Thread {self.id} has crashed, error: {self.response.error}")

		# send stop.
		self.log(f"Stopping thread {self.id}.")
		self.send_stop()
		for i in range(int(timeout/sleeptime)):
			if self.stopped: break
			time.sleep(sleeptime)
		if not self.stopped: 
			return _response_.error(f"Unable to stop thread {self}.")

		# handle stop functions.
		found = False
		try:
			response = self.stop()
			found = True
		except AttributeError: response = None
		if not found:
			try:
				response = self.__stop__()
				found = True
			except AttributeError: response = None
		if isinstance(response, ResponseObject) and not response.success:
			return response

		# handler.
		return _response_.success(f"Successfully stopped thread {self.id}.", required_log_level=0)

		#

	# send stop & crash from within the __run__ function only.
	def send_stop(self, 
		# all optional.
		# option 1: the success message.
		message=None, # (1)
		args={}, # (2)
		# option 2: the error message.
		error=None,
		# option 3: the response object.
		response=None, 
		# save the message/error/response.
		save=False,
		# the active log level (int) (leave None to use self.log_level).
		log_level=None,
		# the required log level for when to print to console (leave None to use _response_.log_level ; default: 0).
		required_log_level=_response_.log_level,
	):
		if log_level == None: log_level = self.log_level
		msg = None
		if isinstance(message, (str,String)):
			self.__response__ = _response_.success(message, args, log_level=-1, save=False)
			msg = message
		elif isinstance(error, (str,String)):
			self.__response__ = _response_.error(error, log_level=-1, save=False)
			msg = error
		elif isinstance(response, ResponseObject):
			self.__response__ = response
			if response["success"]:	
				msg = response["message"]
			else:
				msg = response["error"]
		#else:
		#	raise Exceptions.InvalidUsage("Define one of the following parameters: [message:str, error:str, response:ResponseObject].")
		self.__status__ = "stopped"
		if msg != None and log_level >= required_log_level or save:
			_response_.log(response=self.__response__, save=save, log_level=log_level, required_log_level=required_log_level)
		return self.__response__
		#
	def send_crash(self, 
		# all optional.
		# option 1: the success message.
		message=None, # (1)
		args={}, # (2)
		# option 2: the error message.
		error=None,
		# option 3: the response object.
		response=None, 
		# save the message/error/response.
		save=False,
		# the active log level (int) (leave None to use self.log_level).
		log_level=None,
		# the required log level for when to print to console (leave None to use _response_.log_level ; default: 0).
		required_log_level=_response_.log_level,
	):
		if log_level == None: log_level = self.log_level
		msg = None
		if isinstance(message, (str,String)):
			self.__response__ = _response_.success(message, args, log_level=-1, save=False)
			msg = message
		elif isinstance(error, (str,String)):
			self.__response__ = _response_.error(error, log_level=-1, save=False)
			msg = error
		elif isinstance(response, ResponseObject):
			self.__response__ = response
			if response["success"]:
				msg = response["message"]
			else:
				msg = response["error"]
		#else:
		#	raise Exceptions.InvalidUsage("Define one of the following parameters: [message:str, error:str, response:ResponseObject].")
		self.__status__ = "crashed"
		if msg != None and log_level >= required_log_level or save:
			_response_.log(response=self.__response__, save=save, log_level=log_level, required_log_level=required_log_level)
		return self.__response__
		#

		#

	# copy functions.
	def log(self, 
		# option 1:
		# the message (#1 param).
		message=None,
		# option 2:
		# the error.
		error=None,
		# option 3:
		# the response dict (leave message None to use).
		response={},
		# print the response as json.
		json=JSON,
		# optionals:
		# the active log level (leave None to use self.log_level).
		log_level=None,
		# the required log level for when printed to console.
		required_log_level=0, 
		# save to log file.
		save=False,
		# save errors always (for options 2 & 3 only).
		save_errors=None,
		# the log mode (leave None for default).
		mode=None,
	):
		if log_level == None: log_level = self.log_level
		return _response_.log( 
			# option 1:
			# the message (#1 param).
			message=message,
			# option 2:
			# the error.
			error=error,
			# option 3:
			# the response dict (leave message None to use).
			response=response,
			# print the response as json.
			json=json,
			# optionals:
			# the active log level.
			log_level=log_level,
			# the required log level for when printed to console (leave None to use self.log_level).
			required_log_level=required_log_level, 
			# save to log file.
			save=save,
			# save errors always (for options 2 & 3 only).
			save_errors=save_errors,
			# the log mode (leave None for default).
			mode=mode,
		)

	#properties
	@property
	def run_permission(self):
		return self.__status__ in ["running"]
	@property
	def running(self):
		return self.__status__ in ["running"]
	@property
	def stopped(self):
		return self.__status__ in ["stopped"]
	@property
	def crashed(self):
		return self.__status__ in ["crashed"]
	@property
	def response(self):
		return self.__response__
	# str representation.
	def __str__(self):
		error = None
		if self.crashed:
			try: error = self.response.error
			except: error = None
		if error != None:
			return f"<{self.__traceback__()[1:-1]} (crashed: {self.crashed}) (error: {error}) >"
		else:
			return f"<{self.__traceback__()[1:-1]} (running: {self.running}) (stopped: {self.stopped}) (crashed: {self.crashed}) >"
	# representation.
	def __repr__(self):
		return str(self)

		#

	#
	
#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.response import *
from dev0s.classes.defaults import objects
from dev0s.classes.response import response as _response_

# pip imports.
import urllib
import requests as __requests__

# the requests class.
class Requests(object):
	def __init__(self):
		
		# docs.
		DOCS = {
			"module":"dev0s.requests", 
			"initialized":True,
			"description":[], 
		}

		# attributes.
		self.https = True
		self.allow_redirects = True

	# encode & decode data.
	def encode(self, data={}):
		_data_ = {}
		for key,value in data.items():
			if value.__class__.__name__ in ["OutputObject", "ResponseObject"]:
				value = str(value.dict())
			elif isinstance(value, (list,Array,dict,Dictionary)):
				value = str(value)
			_data_[key] = value
		return f"?{urllib.parse.urlencode(_data_)}"
	def quote(self, data={}):
		return urllib.parse.quote(json.dumps(data))
	def unquote(self, encoded, depth=30):
		return json.loads(urllib.parse.unquote(encoded))

	# get.
	def get(self,
		# the url (str) (#1).
		url=None,
		# the sended post data (dict) (#2).
		data={},
		# serialize output to dictionary.
		serialize=False,
	):
		
		# url.
		url = url.replace("http://", "").replace("https://", "")
		url = f"{Boolean(self.https).string(true='https', false='http')}://"+gfp.clean(f"{url}/", remove_double_slash=True, remove_last_slash=False, remove_first_slash=True)
		if data != {}: url += self.encode(data)

		# request.
		original_request_object = __requests__.get(url, allow_redirects=self.allow_redirects)
		if original_request_object.status_code != 200:
			return _response_.error(f"Invalid request ({url}) [{original_request_object.status_code}]: {original_request_object.text}")
		if serialize:
			try: response = _response_.ResponseObject(original_request_object.json())
			except Exception as e: 
				return _response_.error(f"Request ({url}) [{original_request_object.status_code}]: Unable to serialize output: {original_request_object.text}.")
			return response
		return original_request_object


	# the restapi over ssh object class.
	class RestAPI(objects.Object):
		def __init__(self, 
			# the root url (optional).
			url=None,
			# the default data send with every request (will be appended to local data).
			data={
				"api_key":None,
			},
		):
			
			# docs.
			DOCS = {
			"module":"dev0s.requests.RestAPI", 
			"initialized":False,
			"description":[], }

			# defaults.
			objects.Object.__init__(self, traceback="dev0s.requests.RestAPI")
			
			# attributes.
			self.url = url
			self.data = data

			#
		def request(self, url="/", data={}):
			
			# data.
			for key,value in self.data.items(): 
				data[key] = value

			# url.
			if self.url != None: 
				url = gfp.clean(f"{self.url}/{url}/", remove_double_slash=True, remove_last_slash=False)
			
			# request.
			return requests.get(serialize=True, url=url, data=data)

			#

		#

	#

# initialized classes.
requests = Requests()

#

#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# exceptions:
class Exceptions():
	class JSONDecodeError(Exception):
		pass
	class DuplicateError(Exception):
		pass
	class InstanceError(Exception):
		pass
	class FormatError(Exception):
		pass
	class UnpackError(Exception):
		pass
	class InvalidUsage(Exception):
		pass
	class InvalidFormatError(Exception):
		pass
	class ArgumentFormatError(Exception):
		pass
	class EmptyArgumentError(Exception):
		pass
	class ArgumentError(Exception):
		pass
	class UnknownArgumentError(Exception):
		pass
	class NetworkError(Exception):
		pass
	class InvalidOperatingSystem(Exception):
		pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.defaults.files import *
from dev0s.classes.response import response as _response_
from dev0s.classes import code
from dev0s.classes.defaults.defaults import defaults

"""
missing imports / variables:
	defaults, python_plus
"""

# functions.
def check_user(id, create=False):
	user = User(username=id)
	response = user.check()
	if not response.success:  return response
	if response["success"] and not response["exists"] and create: 
		response = user.create()
		if not response.success: return response
	return _response_.success(f"Successfully checked user [{id}].")
def check_group(id, users=[], create=False, overwrite=False):
	group = Group(name=id, get_users=False)
	response = group.check()
	if not response.success: return response
	if response["success"] and not response["exists"] and create: 
		response = group.create(users=users)
		if not response.success: return response
	elif overwrite:
		response = group.check_users(users=users)
		if not response.success: return response
	return _response_.success(f"Successfully checked group [{id}].")


# the user object.
class User(object):
	# notes about the object
	# this is an example
	def __init__(self, 
		# the users username.
		username=None,
	):

		# docs.
		DOCS = {
			"module":"dev0s.system.User", 
			"initialized":False,
			"description":[], }

		# arguments.
		self.username = username
		self.home_directory = f"{defaults.vars.homes}{self.username}/"

		#
	def create(self):
		
		# check duplicates.
		l_response = self.check()
		if l_response["error"] != None: return l_response
		if l_response["exists"]:
			return _response_.error(f"User [{self.username}] already exists.")

		# check home dir.
		if Files.exists(self.home_directory): 
			return _response_.error(f"Home directory [{self.home_directory}] already exists.")

		# handle linux.
		if OS in ["linux"]:
			# ubuntu.
			output = code.execute(f"sudo useradd -s /bin/bash -m {self.username}")

			# success.
			if output == "":
				return _response_.success(f"Successfully created user [{self.username}].")

			else:
				return _response_.error(String(output.output.replace("useradd: ", "").replace("\n", ". ")).capitalized_word())

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		
		

		#
	def delete(self):
		
		# check existance.
		l_response = self.check()
		if l_response["error"] != None: return l_response
		if not l_response["exists"]:
			return _response_.error(f"User [{self.username}] does not exist.")

		# handle linux.
		if OS in ["linux"]:

			# delete.
			os.system(f"sudo userdel -r {self.username}")
			os.system(f"sudo rm -fr {self.home_directory}")

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		# check.
		l_response = self.check()
		if l_response["error"] != None: return l_response
		
		# success.
		if not l_response["exists"]:
			return _response_.success(f"Successfully deleted user [{self.username}].")
		else:
			return _response_.error(f"Failed to delete user [{self.username}].")

		#
	def check(self, silent=False):

		# handle linux.
		exists = False
		if OS in ["linux"]:
			try: 
				output = subprocess.check_output("sudo id {username}".format(username=self.username), shell=True).decode()
				exists = True
			except: exists = False

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		# success.
		return _response_.success(f"Successfully checked user [{self.username}].", {
			"exists":exists,
		})

		#
	def set_password(self, password=None):
		
		# check params.
		success, response = _response_.parameters.check(
			parameters={
				"password":password,
			})
		if not success: return response

		# handle linux.
		if OS in ["linux"]:

			# get output.
			output = code.execute(f"export HISTIGNORE='*sudo -S*' && echo '{password}\n{password}\n' | sudo -S -k sudo passwd {self.username}")
			
			# success.
			if "password updated successfully" in output:
				return _response_.success(f"Successfully edited the password of user [{self.username}].")

			# fail.
			else:
				print(output)
				return _response_.error(f"Failed to edit the password of user [{self.username}].")

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		#
	def add_groups(self, groups=[]):

		# iterate groups.
		for group in groups:

			# init group.
			group = GroupObject(name=group)

			# check existance.
			l_response = group.check()
			if l_response["error"] != None: return l_response
			if not response["exists"]: 
				return _response_.error(f"Group [{group.name}] does not exist.")

			# add user.
			l_response = group.add_users(users=[self.username])
			if l_response["error"] != None: return l_response

			#
		
		# success.
		return _response_.success(f"Successfully added user to [{len(groups)}] groups.")

		#
	def delete_groups(self, groups=[]):
		
		# iterate groups.
		for group in groups:
			group = GroupObject(name=group)
			l_response = group.check()
			if l_response["error"] != None: return l_response
			if not response["exists"]: 
				return _response_.error(f"Group [{group.name}] does not exist.")

			# delete user.
			l_response = group.delete_users(users=[self.username])
			if l_response["error"] != None: return l_response

				
		# success.
		return _response_.success(f"Successfully added user to [{len(groups)}] groups.")

		#

# the group object.
class Group(object):
	def __init__(
		self, 
		# string format.
		name=None,
		users=[], # all authorized user identifiers.
		# boolean format.
		get_users=False, # (only gets filled if the storages group exists.)
	):

		# docs.
		DOCS = {
			"module":"dev0s.system.Group", 
			"initialized":False,
			"description":[], }

		# arguments.
		self.name = name
		self.users = users

		# functions.
		if get_users:
			response = self.check()
			if response["success"] and response["exists"]:
				response = self.list_users()
				if response["success"]: self.users = response["users"]

		#
	def create(self, users=None):

		# initialize default response.
		if users == None: users = self.users

		# check existance.
		l_response = self.check()
		if l_response["error"] != None: return l_response
		elif l_response["exists"]:
			return _response_.error(f"Group [{self.name}] already exists.")

		# handle linux.
		if OS in ["linux"]:
			output = code.execute(f"sudo groupadd {self.name}",)

			# success.
			if output == "":
				return _response_.success(f"Successfully created group [{self.name}].")

			else:
				return _response_.error(String(output.output.replace("groupadd: ", "").replace("\n", ". ")).capitalized_word())

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		#
	def delete(self):

		# check existance.
		l_response = self.check()
		if l_response["error"] != None: return l_response
		elif l_response["exists"]:
			return _response_.error(f"Group [{self.name}] already exists.")

		# handle linux.
		if OS in ["linux"]:
			os.system(f"sudo groupdel {self.name}")

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		# check existance.
		l_response = self.check()
		if l_response["error"] != None: return l_response
		elif l_response["exists"]:
			return _response_.error(f"Failed to delete group [{self.name}].")
		else:
			return _response_.success(f"Successfully deleted group [{self.name}].")


		#
	def check(self):

		# handle linux.
		exists = False
		if OS in ["linux"]:
			try:
				output = subprocess.check_output("grep '^{name}' /etc/group".format(name=self.name), shell=True).decode()
				exists = True
			except: exists = False

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		
		# success.
		return _response_.success(f"Successfully checked group [{self.name}].", {
			"exists":exists,
		})

		#
	def list_users(self):

		# check existance.
		l_response = self.check()
		if l_response["error"] != None: return l_response
		elif not l_response["exists"]:
			return _response_.error(f"Group [{self.name}] does not exists.")

		# handle linux.
		users = []
		if OS in ["linux"]:
			try: output = subprocess.check_output("members "+self.name, shell=True).decode().replace("\n", "").split(" ")
			except: output = []
			for i in output:
				if i not in [""]: 
					users.append(i.replace("\n", ""))

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		# success.
		self.users = users
		return _response_.success(f"Successfully listed all users {len(users)} of group [{self.name}].", {
			"users":users,
		})

		#
	def delete_users(self, users=[]):

		# check existance.
		l_response = self.check()
		if l_response["error"] != None: return l_response
		elif not l_response["exists"]:
			return _response_.error(f"Group [{self.name}] does not exists.")

		# handle linux.
		if OS in ["linux"]:
			for user in users:
				output = code.execute(f"sudo deluser {user} {self.name}")
				if output != "":
					return _response_.error(String(output.output.replace("deluser: ", "").replace("\n", ". ")).capitalized_word())

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		# success.
		return _response_.success(f"Successfully deleted {len(users)} users from group [{self.name}].", {
			"users":users,
		})

		#
	def add_users(self, users=[]):

		# check existance.
		l_response = self.check()
		if l_response["error"] != None: return l_response
		elif not l_response["exists"]:
			return _response_.error(f"Group [{self.name}] does not exists.")

		# handle linux.
		if OS in ["linux"]:
			for user in users:
				output = code.execute(f"sudo usermod -a -G {self.name} {user}")
				if not output.success: return output
				if output.output != "":
					return _response_.error(String(output.output.replace("usermod: ", "").replace("\n", ". ")).capitalized_word())

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		# success.
		return _response_.success(f"Successfully added {len(users)} users to group [{self.name}].", {
			"users":users,
		})

		#
	def check_users(self, users=[]):
		# deletes all users that are not in the specified ones & adds new specified ones.
		#

		# check existance.
		l_response = self.check()
		if l_response["error"] != None: return l_response
		elif not l_response["exists"]:
			return _response_.error(f"Group [{self.name}] does not exists.")

		# handle linux.
		to_delete, to_add  = [], []
		if OS in ["linux"]:
			
			# check to delete:
			response = self.list_users()
			if response["error"] != None: return response
			l_users = response["users"]
			for user in l_users:
				if user not in users: to_delete.append(user)
			if len(to_delete) > 0:
				response = self.delete_users(users=to_delete)
				if response["error"] != None: return response

			# check to add:
			response = self.list_users()
			if response["error"] != None: return response
			l_users = response["users"]
			for user in users:
				if user not in l_users: to_add.append(user)
			if len(to_add) > 0:
				response = self.add_users(users=to_add)
				if response["error"] != None: return response

		# handle macos.
		elif OS in ["macos"]:
			return _response_.error(f"Unsupported operating system [{OS}].")

		# success.
		return _response_.success(f"Successfully added {len(to_add)} & removed {len(to_delete)} users from group [{self.name}].")

		#

"""
# the unix manager.
class UnixManager(object):
	# notes about the object
	# this is an example
	def __init__(self):

		# init:
		self.users_path = "/home/"
		if OS in ["macos"]: self.users_path = "/Users/"
		self.users = {} # can be filles with objects, format = {"$idenfitier":UserObject()}
		self.groups = {} # can be filles with objects, format = {"$idenfitier":GroupObject()}

"""






"""

# -----------------------
# the User() object class.

# initialize a user object.
user = User("testuser")

# check if the user exists.
response = user.check()
if response["success"]: print("User existance:",response["exists"])

# create a user.
response = user.create()

# delete a user.
response = user.delete()

# set a users password.
response = user.set_password(password="Doeman12!")

# add the user to groups.
response = user.add_groups(groups=[])

# delete the user from groups.
response = user.add_groups(groups=[])

# -----------------------
# the Group() object class.

# initialize a group object.
group = Group("testgroup")

# check if the group exists.
response = group.check()
if response["success"]: print("Group existance:",response["exists"])

# create a group.
response = group.create()

# delete a group.
response = group.delete()

# list the current users.
response = group.list_users()
if response["success"]: print(f"Users of group {group.name}:",response["users"])

# add users to the group.
response = group.add_users(users=["testuser"])

# delete users from the group.
response = group.delete_users(users=["testuser"])

# check if specified users are enabled and remove all other users.
response = group.check_users(users=["testuser"])

# -----------------------
# The dictionary response.

# handle response.
if response["error"] != None: print(response["error"])
else: print(response["message"])

"""


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.config import *
from dev0s.classes.defaults.defaults import defaults
from dev0s.classes.response import response as _response_
from dev0s.classes.defaults.objects import *
from dev0s.classes import code

# the service object class.
class Service(Object):
	# launchd: 
	# 	$ tail -f /var/log/system.log
	#	https://medium.com/swlh/how-to-use-launchd-to-run-services-in-macos-b972ed1e352
	#	https://davidhamann.de/2018/03/13/setting-up-a-launchagent-macos-cron/
	def __init__(self,
		# the service id.
		id=None,
		# the user & group on which the service will be run.
		user=None,
		group=None,
		# the start command.
		start=None,
		# the service description.
		description="",
		# restart on crash.
		restart=True,
		# the restart limit.
		restart_limit=5,
		# the restart delay.
		restart_delay=10,
		# the path to the log file.
		logs=None,
		# the path to the error file.
		errors=None,
		# the object's log level.
		log_level=0,
		# the import traceback.
		traceback="dev0s.system.Service",
	):	

		# docs.
		DOCS = {
			"module":"dev0s.system.Service", 
			"initialized":False,
			"description":[], }

		# defaults.
		Object.__init__(self, traceback=traceback)

		# vars.
		self.id = id
		self.user = user
		self.group = group
		self.start_ = start
		self.description = description
		self.restart_limit = restart_limit
		self.restart_delay = restart_delay
		self.log_level = log_level
		self.options = []
		self.logs, self.errors = None, None
		if restart:
			self.options.append("restart_")
		if logs != None: 
			self.options.append("logs")
			self.logs = File(path=logs)
			if not Files.exists(path=self.logs.fp.base()):
				os.system(f"mkdir -p {self.logs.fp.base()}")
			if self.logs.fp.exists(): self.logs.load()
		if errors != None: 
			self.options.append("errors")
			self.errors = File(path=errors)
			if not Files.exists(path=self.errors.fp.base()):
				os.system(f"mkdir -p {self.errors.fp.base()}")
			if self.errors.fp.exists(): self.errors.load()

		# attributes.
		if defaults.vars.os in ["macos"]:
			self.service = File(path=f"/Library/LaunchDaemons/{self.id}.plist")
		else:
			self.service = File(path=f"/etc/systemd/system/{self.id}.service")

		#
	# functions.
	def create(self):

		# check params.
		response = _response_.parameters.check(
			traceback=self.__traceback__(function="create"),
			parameters={
				"id":self.id,
				"user":self.user,
				"start":self.start_,
			})
		if not response.success: return response

		# checks.
		if self.service.fp.exists(): 
			return _response_.error(f"Service [{self.service.fp.path}] already exists (call service.check() instead).")

		# save.
		_response_.log(f"&ORANGE&Root permission&END& required to save changes to [{self.service.fp}].", log_level=self.log_level)
		self.service.save(data=self.__create__(), sudo=True)
		self.service.fp.ownership.set("root", sudo=True)
		self.service.fp.permission.set(700, sudo=True)
		os.system("sudo systemctl daemon-reload")

		# handler.
		if self.service.fp.exists(): 
			return _response_.success(f"Successfully created service {self.id}.")
		else:
			return _response_.error(f"Failed to create service {self.id}.")

		#
	def check(self):

		# check params.
		response = _response_.parameters.check(
			traceback=self.__traceback__(function="check"),
			parameters={
				"id":self.id,
				"user":self.user,
				"start":self.start_,
			})
		if not response.success: return response
		
		# checks.
		if not self.service.fp.exists(): 
			return self.create()

		# save.
		self.service.load(sudo=True)
		data = self.__create__()
		if self.service.data != data:
			_response_.log(f"&ORANGE&Root permission&END& required to save changes to [{self.service.fp}].", log_level=self.log_level)
			self.service.save(data=data, sudo=True)
			os.system("sudo systemctl daemon-reload")

		# always set permissions.
		self.service.fp.ownership.set("root", sudo=True)
		self.service.fp.permission.set(700, sudo=True)
		
		# handler.
		if self.service.fp.exists(): 
			return _response_.success(f"Successfully checked service {self.id}.")
		else:
			return _response_.error(f"Failed to check service {self.id}.")

		#
	def delete(self):

		# checks.
		if not self.service.fp.exists(): 
			return _response_.error(f"Service [{self.service.fp.path}] does not already exist.")

		# save.
		self.service.fp.delete(sudo=True, forced=True)

		# handler.
		if not self.service.fp.exists(): 
			return _response_.success(f"Successfully deleted service {self.id}.")
		else:
			return _response_.error(f"Failed to delete service {self.id}.")

		#
	# control functions.
	def start(self):
		
		# launchd.
		if defaults.vars.os in ["macos"]:
			command = ""
			for i in [self.id]:
				command += f"sudo launchctl start {i} &&"
			command = command[:-3]
			output = code.execute(command)
			if not output.success: return output
			elif output.output in ["", "\n"]:
				return _response_.success(f"Successfully started {self.id}.")
			else:
				return _response_.error(f"Failed to start {self.id}.")

		# systemd.
		else:
			command = ""
			for i in [self.id]:
				command += f"sudo systemctl start {i} &&"
			command = command[:-3]
			output = code.execute(command)
			if not output.success: return output
			elif output.output in ["", "\n"]:
				return _response_.success(f"Successfully started {self.id}.")
			else:
				return _response_.error(f"Failed to start {self.id}.")

		#
	def stop(self):
		
		# launchd.
		if defaults.vars.os in ["macos"]:
			command = ""
			for i in [self.id]:
				command += f"sudo launchctl stop {i} &&"
			command = command[:-3]
			output = code.execute(command)
			if not output.success: return output
			elif output.output in ["", "\n"]:
				return _response_.success(f"Successfully stopped {self.id}")
			else:
				return _response_.error(f"Failed to stop {self.id}")

		# systemd.
		else:
			command = ""
			for i in [self.id]:
				command += f"sudo systemctl stop {i} &&"
			command = command[:-3]
			output = code.execute(command)
			if not output.success: return output
			elif output.output in ["", "\n"]:
				return _response_.success(f"Successfully stopped {self.id}")
			else:
				return _response_.error(f"Failed to stop {self.id}")

		#
	def restart(self):
		
		# launchd.
		if defaults.vars.os in ["macos"]:
			response = self.stop()
			if not response.success: return response
			response = self.start()
			if not response.success: return response
			return _response_.success(f"Successfully restarted {self.id}")

		# systemd.	
		else:
			command = ""
			for i in [self.id]:
				command += f"sudo systemctl restart {i} &&"
			command = command[:-3]
			output = code.execute(command)
			if not output.success: return output
			elif output.output in ["", "\n"]:
				return _response_.success(f"Successfully restarted {self.id}")
			else:
				return _response_.error(f"Failed to restarted {self.id}")

		#
	def status(self):
		
		# launchd.
		if defaults.vars.os in ["macos"]:
			output = code.execute(f"sudo launchctl list | grep {self.id}")
			if not output.success: return output
			else:
				return _response_.success(f"Successfully retrieved the status of {self.id}",{
					"status":output.output,
				})

		# systemd.
		else:
			#output = code.execute(f"sudo systemctl status {self.id}")
			output = code.execute(f"sudo systemctl status {self.id} > /tmp/status && cat /tmp/status && rm -fr /tmp/status")
			if not output.success: return output
			else:
				return _response_.success(f"Successfully retrieved the status of {self.id}",{
					"status":output.output,
				})

			#
	def reset_logs(self):

		# execute.
		if "logs" in options:
			self.logs.save(data="")
		if "errors" in options:
			self.logs.save(data="")
		return _response_.success("Successfully resetted the logs.")

		#
	def tail(self, global_=False, debug=False):
			
		# load logs.
		if defaults.vars.os in ["macos"] and global_:
				data = code.execute(f"cat /var/log/system.log | grep {self.id}")
				if not data.success: return data
				data = str(data)
		elif debug:
			data = self.logs.load()
		else:
			data = self.errors.load()
		
		# handler.
		return _response_.success(f"Successfully tailed the {self.id} logs.", {
			"logs":data,
		})

		#
	# system functions.
	def __create__(self):
		
		# lauchd.
		if defaults.vars.os in ["macos"]:

			# checks.
			if not isinstance(self.start_, (list, Array)): 
				raise ValueError(f"{self.traceback} attribute [start] requires to be a [list] or [Array] not [{self.start_.__class__.__name__}].")

			# vars.
			arguments, restart, logs, errors, group = "", "", "", "", ""

			# arguments.
			if len(self.start_) >= 0:
				arguments =  f'    <key>ProgramArguments</key>' + '\n'
				arguments += f'    <array>' + '\n'
				for i in self.start_:
					arguments += f'        <string>{i}</string>' + '\n'
				arguments += f'    </array>' + '\n'

			# group.
			if self.group != None:
				group = (
					"    <key>GroupName</key>" + "\n" +
					f"    <string>{self.group}</string>" + "\n" +
					"")

			# restart.
			if "restart" in self.options:
				restart = (
					f'    <key>StartInterval</key>' + '\n' +
					f'    <integer>{self.restart_delay}</integer>' + '\n' +
					"")

			# logs.
			if "logs" in self.options:
				logs = (
					f"    <key>StandardOutPath</key>" + "\n" +
					f"    <string>{self.logs.fp.path}</string>" + "\n" +
					"")

			# errors.
			if "errors" in self.options:
				errors = (
					"    <key>StandardErrorPath</key>" + "\n" +
					f"    <string>{self.errors.fp.path}</string>" + "\n" +
					"")

			# combine.
			service = (
				f'<?xml version="1.0" encoding="UTF-8"?>' + '\n' +
				f'<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">' + '\n' +
				f'<plist version="1.0">' + '\n' +
				f'<dict>' + '\n' +
				f'    <key>Label</key>' + '\n' +
				f'    <string>{self.id}</string>' + '\n' +
				f"    <key>UserName</key>" + "\n" +
				f"    <string>{self.user}</string>" + "\n" +
				group + 
				arguments +
				restart +
				logs +
				errors + 
				f'</dict>' + '\n' +
				f'</plist>' + '\n' +
				'')
			
		# systemd.	
		else:

			# checks.
			if isinstance(self.start_, (list, Array)):
				start = Array(self.start_).string(joiner=" ")
			else:
				start = str(self.start_)

			# build.
			service = (
				f"[Unit]" + "\n" +
				f"Description={self.description}" + "\n" +
				f"After=network.target" + "\n" +
				f"StartLimitIntervalSec=0" + "\n" +
				f"" + "\n" +
				f"[Service]" + "\n" +
				f"Type=simple" + "\n" +
				f"ExecStart={start}" + "\n" +
				f"User={self.user}" + "\n" +
				f""
			)

			# group.
			if self.group != None:
				service += (
					f"Group={self.group}" + "\n" +
					"")

			# restart.
			if "restart" in self.options:
				service += (
					f"Restart=always" + "\n" +
					f"RestartSec=1" + "\n" +
					"")
				if self.restart_limit != None:
					service += ( 
						f"StartLimitBurst={self.restart_limit}" + "\n" +
						f"StartLimitIntervalSec={self.restart_delay}" + "\n" +
						f"")

			# additional build.
			service += (
				f"" + "\n" +
				f"[Install]" + "\n" +
				f"WantedBy=multi-user.target" + "\n" +
				f"")

		# handler.
		if service[len(service)-1] == "\n": service = service[:-1]
		return service

#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.config import *
from dev0s.classes.response import response as _response_
from dev0s.classes import code

# the firewall class.
class FireWall(object):
	def __init__(self):	
		
		# docs.
		DOCS = {
			"module":"dev0s.network.firewall", 
			"initialized":True,
			"description":[], }

		#
	def enable(self):

		# check os.
		response = self.__check_os__()
		if response["error"] != None: return response

		# do.
		output = code.execute("printf 'y\\ny\\ny\\ny\\ny\\ny\\n' | sudo ufw enable")
		if "Firewall is active and enabled on system startup" in output:
			return _response_.success(f"Successfully enabled the firewall.")
		else:
			return _response_.error(f"Failed to enable the firewall.")

		#
	def disable(self):
		
		# check os.
		response = self.__check_os__()
		if response["error"] != None: return response

		# do.
		output = code.execute("printf 'y\\ny\\ny\\ny\\ny\\ny\\n' | sudo ufw disable")
		if "Firewall stopped and disabled on system startup" in output:
			return _response_.success(f"Successfully disabled the firewall.")
		else:
			return _response_.error(f"Failed to disable the firewall.")

		#
	def allow(self, port):
		

		# check os.
		response = self.__check_os__()
		if response["error"] != None: return response

		# do.
		output = utils.__execute__(["sudo", "ufw", "allow", str(port)])
		if "Rule updated" in output or "Rules updated" in output or "Skipping adding existing rule" in output:
			return _response_.success(f"Successfully allowed port [{port}].")
		else:
			return _response_.error(f"Failed to allow port [{port}], output: [{output}].")

		#
	def deny(self, port):
		

		# check os.
		response = self.__check_os__()
		if response["error"] != None: return response

		# do.
		output = utils.__execute__(["sudo", "ufw", "deny", str(port)])
		if "Rule updated" in output or "Rules updated" in output or "Skipping adding existing rule" in output:
			return _response_.success(f"Successfully denied port [{port}].")
		else:
			return _response_.error(f"Failed to deny port [{port}], output: [{output}].")

		#
	def allow_all(self):
		response = self.info()
		if response["error"] != None: return response
		response = self.set_default(deny=True)
		if response["error"] != None: return response
		ports = response["ports"]
		for port, info in ports.items():
			response = self.allow(port)
			if response["error"] != None: return response
		return _response_.success(f"Successfully allowed all {len(ports)} ports.")
	def deny_all(self):
		response = self.info()
		if response["error"] != None: return response
		response = self.set_default(deny=False)
		if response["error"] != None: return response
		ports = response["ports"]
		for port, info in ports.items():
			response = self.allow(port)
			if response["error"] != None: return response
		return _response_.success(f"Successfully allowed all {len(ports)} ports.")
	def set_default(self, deny=True):
		

		# check os.
		response = self.__check_os__()
		if response["error"] != None: return response

		# do.
		option = None
		if deny: option = "deny"
		else: option = "allow"
		output = utils.__execute__(["sudo", "ufw", "default", option])
		if "Default incoming policy changed to " in output:
			return _response_.success(f"Successfully set the default action to [{option}].")
		else:
			return _response_.error(f"Failed set the default action to [{option}].")

		#
	def info(self):
		def __handle_port_line__(line):
			new = []
			for i in line.split("  "):
				if i not in ["", " "]:
					new.append(i)
			_new_ = []
			for i in new:
				for a in range(101):
					if len(i) > 0 and i[0] == " ": i = i[1:]
					elif len(i) > 0 and i[len(i)-1] == " ": i = i[:-1]
					else: break
				_new_.append(i)
			return {
				"port":_new_[0],
				"status":_new_[1],
				"from":_new_[2],
			}

		# check os.
		response = self.__check_os__()
		if response["error"] != None: return response

		# retrieve.
		output = utils.__execute__(["sudo", "ufw", "status"])
		#print(f"Output: [{output}].")

		# get status.
		status = "unknown"
		if "Status: " in output: 
			try: status = output.split("Status: ")[1].split("\n")[0].replace(" ","")
			except IndexError: status = "unknown"

		# iterate ports.
		ports = {}
		output = output.split('\n')
		if len(output) > 1:
			c, set = 0, 0
			for i in output:
				if 'To' in i and 'Action' in i and 'From' in i:
					set += 1
				elif set >= 2:
					if i not in ["", " "]:
						info = __handle_port_line__(i)
						ports[info["port"]] = info
				elif set != 0: set += 1
				c += 1

		# success.
		return _response_.success("Successfully retrieved the firewall info.", {
			"status":status,
			"ports":ports,
		})

		#
	# system functions.
	def __check_os__(self):
		if OS not in ["linux"]:
			return _response_.error(f"Unsupported operating system: [{OS}].")
		return _response_.success(f"Supported operating system: [{OS}].")

# initialized classes.
firewall = FireWall()

"""

# retrieve the firewall information.
response = firewall.info()

# disable the firewall.
response = firewall.disable()

# enable the firewall.
response = firewall.enable()

# set the default port action.
response = set_default(deny=True)

# allow a port.
response = firewall.allow(2200)

# deny a port.
response = firewall.deny(2200)

"""


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# _____________________________________________________________________
#
# Shortcuts 
#   * imported by import order.
#   * should be imported as.
# 		from dev0s.shortcuts import *
# 
#

# default pip imports.
import os, sys

# import.
import dev0s

# utils.
utils = dev0s.utils

# exceptions.
exceptions = dev0s.exceptions

# color.
color = dev0s.color
symbol = dev0s.symbol

# console.
Loader = dev0s.console.Loader
ProgressLoader = dev0s.console.ProgressLoader

# files & formats.
Files = dev0s.files.Files
Formats = dev0s.files.Formats
gfp = dev0s.files.gfp
gdate = dev0s.files.gdate
Files = dev0s.files.Files
Formats = dev0s.files.Formats
FilePath = Formats.FilePath 
String = Formats.String 
Boolean = Formats.Boolean 
Integer = Formats.Integer 
Date = Formats.Date
Generate = Formats.Generate
Interval = Formats.Interval
File = Files.File
Directory = Files.Directory
Zip = Files.Zip
Image = Files.Image
Bytes = Files.Bytes
Dictionary = Files.Dictionary
Array = Files.Array

# response.
ResponseObject = dev0s.response.ResponseObject

# objects.
Traceback = dev0s.objects.Traceback
Object = dev0s.objects.Object
Thread = dev0s.objects.Thread

# class database.
Database = dev0s.database.Database

# requests.
RestAPI = dev0s.requests.RestAPI

# code.
OutputObject = dev0s.code.OutputObject
Spawn = dev0s.code.Spawn
Version = dev0s.code.Version 
Script = dev0s.code.Script
Python = dev0s.code.Python

#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.encryption import aes, rsa, agent

# the encryption module class.
class Encryption():

	# aes.
	AES = aes.AES
	AsymmetricAES = aes.AsymmetricAES
	Database = aes.Database

	# rsa.
	RSA = rsa.RSA
	#EncryptedDictionary = rsa.EncryptedDictionary

	# agent.
	Agent = agent.Agent

	


#!/usr/bin/env python3

# Note!
# ' are required, do not use any '.

# setup.
from setuptools import setup, find_packages
setup(
	name='dev0s',
	version='3.16.7',
	description='Some description.',
	url='http://github.com/vandenberghinc/dev0s',
	author='Daan van den Bergh',
	author_email='vandenberghinc.contact@gmail.com',
	license='MIT',
	packages=find_packages(),
	zip_safe=False,
	include_package_data=True,
	install_requires=[
            'wheel',
            'django',
            'requests',
            'pathlib',
            'pexpect',
            'certifi',
            'chardet',
            'pycryptodome',
            'urllib3',
            'flask',
            'selenium',
            'jinja2',
            'werkzeug',
            'click',
        ])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.defaults.files import *
from dev0s.classes.defaults.defaults import defaults
from dev0s.classes.response import response as _response_
from dev0s.classes.defaults.objects import *
from dev0s.classes import database as _database_
import getpass

# lib imports.
from dev0s.classes.encryption import aes

# the agent object class.
class Agent(Traceback):
	def __init__(self,
		# the encryption & webserver's id (str).
		id="dev0s-agent",
		# the path to the encrypted database (str, String, FilePath).
		database=None,
		# the webserver's host (str).
		host="127.0.0.1",
		# the webserver's port (int).
		port=56000,
		# the path to the private key / the raw private key (str).
		private_key=None,
		# the path to the public key / the raw public key (str).
		public_key=None,
		# the passphrase (optional to prompt) (str).
		passphrase=None,
		# the encryption key in memory only (enable when you passed the private_key & public_key in raw format and the file path) (bool).
		memory=False,
		# the interactive mode (prompt for password) (bool).
		interactive=True,
		# the object traceback (str).
		traceback="dev0s.encryption.Agent",
	):

		# docs.
		DOCS = {
			"module":"dev0s.encryption.Agent", 
			"initialized":False,
			"description":[], }

		# traceback.
		Traceback.__init__(self, traceback=traceback)

		# checks.
		response = _response_.parameters.check(
			traceback=self.__traceback__(),
			parameters={
				"id:str,String":id,
				"database:str,String,FilePath":database,
				"host:str,String":host,
				"port:str,String,int,Integer":port,
				"private_key:str,String":private_key,
				"public_key:str,String":public_key,
				#"passphrase:str,String":passphrase,
				"memory:bool,Boolean":memory,
				"interactive:bool,Boolean":interactive,
				"traceback:str,String":traceback,
			})
		if not response.success: response.crash()

		# check instances.
		if not Files.exists(str(database)):
			_response_.log(f"{color.orange}Root permission{color.end} required to create database [{database}].")
			Files.create(str(database), sudo=True, permission=700, owner=defaults.vars.owner, group=defaults.vars.group, directory=True)


		# init.
		self.id = str(id)
		self.host = str(host)
		self.port = int(port)
		self.private_key = private_key
		self.public_key = public_key
		self.memory = bool(memory)
		self.passphrase = passphrase
		self.interactive = bool(interactive)
		self.db_path = str(database)

		# vars.
		self._activated = False

		# objects.
		self.webserver = _database_.WebServer(
			id=self.id,
			host=self.host,
			port=self.port,
		)
		self.aes = self.encryption = aes.AsymmetricAES(
			public_key=self.public_key,
			private_key=self.private_key,
			passphrase=self.passphrase,
			memory=self.memory,)
		self.db = self.database = aes.Database(path=str(self.db_path), aes=self.encryption)

		#

	# generate encryption.
	def generate(self,
		# the passphrase (optional to prompt) (str).
		passphrase=None,
		# the verify passphrase (optional).
		verify_passphrase=None,
		# interactive (optional).
		interactive=None
	):

		# checks.
		if passphrase == None: passphrase = self.passphrase
		if interactive == None: interactive = self.interactive
		if passphrase == None:
			if not interactive:
				return _response_.error(self.__traceback__(function="generate")+": Define parameter [passphrase].")
			else:
				passphrase = getpass.getpass(f"Enter the passphrase of the {self.id} encryption:")
		elif len(passphrase) < 8: 
			return _response_.error("The passphrase must contain at least 8 characters.")
		elif passphrase.lower() == passphrase: 
			return _response_.error("The passphrase must contain capital characters.")
		elif (interactive and passphrase != getpass.getpass("Enter the same passphrase:")) or (verify_passphrase != None and passphrase != verify_passphrase): 
			return _response_.error("The passphrase must contain at least 8 characters.")

		# check webserver.
		if not self.webserver.running:
			#if not interactive:
			#	return _response_.error(f"{self.traceback}: The webserver is not running.")
			#else:
			if defaults.options.log_level >= 1: _response_.log(f"{ALIAS}: forking {self.id} webserver.")
			response = self.webserver.fork()
			if not response.success: return response

		# generate.
		self.encryption.rsa.passphrase = passphrase
		response = self.encryption.generate_keys()
		if not response["success"]: 
			return _response_.error(f"Encoutered an error while generating the master encryption key: {response['error']}")
		self.passphrase = passphrase
		if self.memory:
			self.encryption.rsa.private_key = response.private_key
			self.encryption.rsa.public_key = response.public_key
		response = self.encryption.load_keys()
		if not response["success"]: 
			return _response_.error(f"Encoutered an error while activating the {self.id} encryption: {response['error']}")

		# cache.
		response = self.webserver.set(group="passphrases", id="master", data=passphrase)
		if not response["success"]: 
			return _response_.error(f"Encoutered an error while caching the passphrase (#1): {response['error']}")

		# database.
		self.db = aes.Database(path=str(self.db_path), aes=self.encryption)
		response = self.db.activate()
		if not response["success"]: 
			return _response_.error(f"Encoutered an error while activating the encrypted cache: {response['error']}")

		# hander.
		return _response_.success("Successfully generated the encryption.")

		#

	# activate encryption.
	def activate(self,
		# the key's passphrase (optional to retrieve from webserver) (str).
		passphrase=None,
		# interactive (optional) 
		interactive=None,
	):
		if passphrase == None: passphrase = self.passphrase
		if interactive == None: interactive = self.interactive
		new = False
		if passphrase in [False, None, "", "null", "None", "none"]:
			
			# check webserver.
			if not self.webserver.running:
				if not interactive:
					return _response_.error(f"{self.traceback}: The webserver is not running.")
				else:
					if defaults.options.log_level >= 1: _response_.log(f"{ALIAS}: forking {self.id} webserver.")
					response = self.webserver.fork()
					if not response.success: return response

			# get pass.
			response, passphrase = self.webserver.get(group="passphrases", id="master"), None
			if not response.success and "There is no data cached for" not in response["error"]: return response
			elif response["success"]: passphrase = response["data"]
			if passphrase in [False, None, "", "null", "None", "none"]:
				if not interactive:
					return _response_.error(self.__traceback__(function="activate")+": Define parameter [passphrase].")
				else:
					new = True
					passphrase = getpass.getpass(f"Enter the passphrase of the {self.id} encryption:")

		# activate.
		self.encryption.rsa.passphrase = passphrase
		response = self.encryption.load_keys()
		if not response["success"]: 
			return _response_.error(f"Encoutered an error while activating the {self.id} encryption: {response['error']}")
		self.passphrase = passphrase
		self.db.aes.rsa.passphrase = passphrase
		response = self.db.activate()
		if not response["success"]: 
			return _response_.error(f"Encoutered an error while activating the encrypted cache: {response['error']}")

		# chache.
		if new:
			response = self.webserver.set(group="passphrases", id="master", data=passphrase)
			if not response["success"]: 
				return _response_.error(f"Encoutered an error while caching the passphrase (#2): {response['error']}")

		# handler.
		return _response_.success("Successfully activated the encryption.")

		#

	# copy functions (only these!).
	def encrypt(self, string, decode=False):
		self.encryption.encrypt(string, decode=decode)
		#
	def decrypt(self, string, decode=False):
		self.encryption.decrypt(string, decode=decode)
		#

	# properties.
	@property
	def activated(self):
		return self.encryption.activated and self.db.aes.activated
		#
	@property
	def public_key_activated(self):
		return self.encryption.public_key_activated and self.db.aes.public_key_activated
		#
	@property
	def private_key_activated(self):
		return self.encryption.private_key_activated and self.db.aes.private_key_activated
		#
	@property
	def generated(self):
		if self.memory:
			return self.encryption.rsa.private_key != None and self.encryption.rsa.public_key != None
		else:
			return self.encryption.rsa.private_key != None and self.encryption.rsa.public_key != None and Files.exists(self.encryption.rsa.private_key) and Files.exists(self.encryption.rsa.public_key)
		#

	# repr.
	def __repr__(self):
		return f"<{self.traceback} (id: {self.id}) (activated: {self.activated}) (generated: {self.generated}) >"
		#

	#
	



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.config import *
from dev0s.classes import code
from dev0s.classes.defaults.defaults import defaults
from dev0s.classes.response import response as _response_
from dev0s.classes.requests import requests
from dev0s.classes.network import firewall

# pip.
import requests as __requests__

# the network object class.
class Network(object):
	def __init__(self):

		# docs.
		DOCS = {
			"module":"dev0s.network", 
			"initialized":True,
			"description":[], }

		# variables.
		self._info_ = None
		self.api_key = None
		if os.environ.get("IPINFO_API_KEY") not in ["None", "", None]:
			self.api_key = os.environ.get("IPINFO_API_KEY")

		# firewall.
		self.firewall = None
		if defaults.vars.os in ["linux"]:
			self.firewall = firewall.FireWall()

		# sys vars.
		self.__cache__ = {}
		self.__timestamps__ = {}

		#
	def info(self):
		try: self.__cache__["info"]
		except KeyError: self.__cache__["info"] = {}
		if self.__cache__["info"] != {}: return self.__cache__["info"]

		# get info.
		"""
		info = None
		if self.api_key == None: 
			request_obj = __requests__.get('https://ipinfo.io/json')
		else: 
			request_obj = __requests__.get('https://ipinfo.io/json', headers={'Authorization': f'Bearer {self.api_key}'})
		try:
			info = request_obj.json()
		except Exception as e:
			return _response_.error(f"Failed to serialze (https://ipinfo.io/json) [{request_obj.status_code}]: {request_obj.text}.")
		try: 
			error = info["error"]
			if "rate limit exceeded" in str(error["title"]).lower():
				return _response_.error(f"(https://ipinfo.io/json) {error['title']}: {error['message'].replace('  ',' ')} Define environment variable $IPINFO_API_KEY to specify an api key.")
			else:
				return _response_.error(f"(https://ipinfo.io/json) {error['title']}: {error['message'].replace('  ',' ')}")
		except KeyError: a=1
		"""

		# request.
		if self.api_key == None:
			response = requests.get('https://ipinfo.io/json', serialize=True)
		else:
			response = requests.get('https://ipinfo.io/json', data={'Authorization': f'Bearer {self.api_key}'}, serialize=True)

		# handle.
		if "success" in response and not response.success: return response
		elif "error" in response:
			if "rate limit exceeded" in str(response["title"]).lower():
				return _response_.error(f"(https://ipinfo.io/json) {response['title']}: {response['message'].replace('  ',' ')} Define environment variable $IPINFO_API_KEY to specify an api key.")
			else:
				return _response_.error(f"(https://ipinfo.io/json) {response['title']}: {response['message'].replace('  ',' ')}")

		# clean.
		if "readme" in response:
			del response["readme"]

		# set.
		info = response
		try: 
			x = info["ip"]
			del(info["ip"])
			info["public_ip"] = x
		except KeyError: 
			return _response_.error("Unable to fetch netork info.")
		try: info["private_ip"] = self.__get_private_ip__()
		except: info["private_ip"] = "unknown"
		try: info["hostname"] = socket.gethostname()
		except: info["hostname"] = "unknown"

		# success.
		self.__cache__["info"] = _response_.success(f"Successfully retrieved the network information.", info)
		return self.__cache__["info"]

		#
	def convert_dns(self, dns, timeout=1):
		response = self.ping(dns, timeout=timeout)
		if response["error"] != None: return response
		if response["ip"] == None: 
			return _response_.error(f"Failed to convert dns [{dns}].")
		return _response_.success(f"Successfully converted dns [{dns}].", {
			"ip":response["ip"]
		})
	def ping(self, ip, timeout=1):

		# set info.
		info = {
			"ip":None,
			"up":False,
		}

		# execute.
		response = code.execute(f"ping {ip}", async_=True, wait=False)
		if not response.success: return response
		process = response.process
		time.sleep(timeout)
		response = process.kill()
		if not response.success: return response
		response = process.read()
		if not response.success: return response
		output = response.output

		# handle.
		info["dns"] = ip
		try: info["ip"] = output.split(f"PING {ip} (")[1].split("):")[0]
		except: info["ip"] = None
		if "Request timeout for" in output:
			info["up"] = False
		elif " bytes from " in output:
			info["up"] = True
		else:
			info["up"] = None

		# success.
		return _response_.success(f"Successfully pinged [{ip}].", info)

		#
	# system functions.
	def __get_private_ip__(self):
		s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		try:
			# doesn't even have to be reachable
			s.connect(('10.255.255.255', 1))
			ip = s.getsockname()[0]
		except Exception:
			ip = '127.0.0.1'
		finally:
			s.close()
		return ip
		#
	# port in use
	def port_in_use(self, port, host="127.0.0.1"):
		a_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		location = (host, port)
		result_of_check = a_socket.connect_ex(location)
		if result_of_check == 0: in_use = True
		else: in_use = False
		a_socket.close()
		return in_use
		#
	# find free port.
	def free_port(self, start=6080):
		for i in range(10000):
			port = start + i
			if not self.port_in_use(port):
				return _response_.success(f"Successfully found a free port.", {
					"port":port,
				})
		return _response_.error(f"Unable to find a free port.")
		#

# initialized classes.
network = Network()

"""

# get network info.
response = network.info("vandenberghinc.com")

# ping an ip.
response = network.ping("192.168.1.200")

# convert a dns.
response = network.convert_dns("vandenberghinc.com")

"""
